import logging
import numpy as np
import os
import json
from matplotlib import rc
from matplotlib.colors import colorConverter
from matplotlib.patches import Rectangle
import matplotlib.pyplot as plt
from collections import OrderedDict
from astropy.io import fits

from getdist import plots, MCSamples

from beagle_utils import BeagleDirectories, prepare_plot_saving, plot_exists

class PDF(object):

    def __init__(self, params_file, mock_catalogue=None, plot_single_solution=None):

        # Names of parameters, used to label the axes, and whether they are log
        # or not
        with open(params_file) as f:    
            # The use of the "OrderedDict" ensures that the order of the
            # entries in the dictionary reflects the order in the file
            self.adjust_params = json.load(f, object_pairs_hook=OrderedDict)

        self.mock_catalogue = mock_catalogue

        self.single_solutions = None
        if plot_single_solution is not None:
            self.single_solutions = OrderedDict()
            with fits.open(plot_single_solution) as f:
                self.single_solutions['ID'] = f[1].data['ID']
                self.single_solutions['row'] = f[1].data['row_index']

    def plot_triangle(self, ID, 
            params_to_plot=None, 
            suffix=None, 
            replot=False, 
            M_star=False, 
            show=False):
        """ 
        Draw a "triangle plot" with the 1D and 2D posterior probability

        Parameters
        ----------
        ID : str or int
            Object ID


        M_star : bool, optional
            If set, the routine will plot the mass currenly locked into stars
            instead of the mass of star formed (i.e., the plotted mass will
            accout for the return fraction)

        """ 
        # NB: you changed the getdist/plot.py _set_locator function
        # replacing line 1172-1176
        #if x and (abs(xmax - xmin) < 0.01 or max(abs(xmin), abs(xmax)) >= 1000):
        #    axis.set_major_locator(plt.MaxNLocator(self.settings.subplot_size_inch / 2 + 3, prune=prune))
        #else:
        #    axis.set_major_locator(plt.MaxNLocator(self.settings.subplot_size_inch / 2 + 4, prune=prune))
        # with
        #if x and (abs(xmax - xmin) < 0.01 or max(abs(xmin), abs(xmax)) >= 1000):
        #    axis.set_major_locator(plt.MaxNLocator(self.settings.subplot_size_inch / 2 + 2, prune=prune))
        #else:
        #    axis.set_major_locator(plt.MaxNLocator(self.settings.subplot_size_inch / 2 + 3, prune=prune))
        # to have a fewer number of tick marks, and hence less crowded triangle plots

        # for "cosmetics" reasons you also changed the getdist/plot.py setWithSubplotSize function
        # replacing line 166-167
        # self.lab_fontsize = 7 + 2 * self.subplot_size_inch
        # self.axes_fontsize = 4 + 2 * self.subplot_size_inch
        # with
        # self.lab_fontsize = 7 + 4 * self.subplot_size_inch
        # self.axes_fontsize = 4 + 4 * self.subplot_size_inch
        # to have larger, hence more readable, label and axes font sizes
    
        # Name of the output plot
        if suffix is None:
            plot_name = str(ID)+'_BEAGLE_triangle.pdf'
        else:
            plot_name = str(ID)+'_BEAGLE_triangle_' + suffix + '.pdf'

        # Check if the plot already exists
        if plot_exists(plot_name) and not replot and not show:
            logging.warning('The plot "' + plot_name + '" already exists. \n Exiting the function.')
            return

        fits_file = os.path.join(BeagleDirectories.results_dir,
                str(ID) + '_' + BeagleDirectories.suffix + '.fits.gz')

        hdulist = fits.open(fits_file)

        param_values = OrderedDict()
        for key, value in self.adjust_params.iteritems():
            extName = "POSTERIOR PDF"
            if "extName" in value:
                extName = value["extName"]

            colName = key
            if "colName" in value:
                colName = value["colName"]

            param_values[key] = hdulist[extName].data[colName]

        probability = hdulist['posterior pdf'].data['probability']

        n_rows = probability.size

       # ParamsToPlot = ['mass', 'redshift', 'tauV_eff', 'metallicity', 'specific_sfr', 'tau']

        # By default you plot all parameters
        if params_to_plot is None:
            _params_to_plot = list()
            for key, value in self.adjust_params.iteritems():
                _params_to_plot.append(key)
        else: 
            _params_to_plot = params_to_plot

        # Here you check whether you want to plot the mass currently locked
        # into stars or not (i.e. accounting for the return fraction as well)
        if M_star and 'mass' in _params_to_plot:
            param_values['mass'][:] = np.log10(hdulist['galaxy properties'].data['M_star'][:])

        nParamsToPlot = len(_params_to_plot)

        names = list()
        labels = list()
        ranges = dict()
        samps = np.zeros((n_rows, nParamsToPlot))
        keys = list()

        j = 0
        for key, par in self.adjust_params.iteritems():
            keys.append(key)
            for par_name in _params_to_plot:
                if key == par_name:
                    names.append(key)
                    label = par['label'].replace("$","")
                    labels.append(label)

                    samps[:,j] = param_values[key]
                    ranges.update({key:par['range']})
                    if 'log' in par:
                        if par["log"]:
                            samps[:,j] = np.log10(param_values[key])
                            ranges.update({key:np.log10(par['range'])})
                    j += 1
                    break

        settings = {
                    "contours":[0.68, 0.95, 0.99], 
                    "range_ND_contour":1, 
                    "range_confidence":0.001,
                    "fine_bins":200,
                    "fine_bins_2d":80,
                    "smooth_scale_1D":0.3,
                    "smooth_scale_2D":0.5,
                    "tight_gap_fraction":0.15
                    }

        samples = MCSamples(samples=samps, names=names, ranges=ranges, \
                weights=probability, labels=labels, settings=settings )

        g = plots.getSubplotPlotter()
        g.settings.num_plot_contours = 3
        g.settings.prob_y_ticks = True

        line_args = {"lw":2, "color":colorConverter.to_rgb("#006FED") } 

        g.triangle_plot(samples, filled=True, line_args=line_args)

        g.fig.subplots_adjust(wspace=0.1, hspace=0.1)

        for i, ax in enumerate([g.subplots[i,i] for i in range(nParamsToPlot)]):
            ax.set_autoscalex_on(True)

        # Add tick labels at top of diagonal panels
        for i, ax in enumerate([g.subplots[i,i] for i in range(nParamsToPlot)]):
            par_name = keys[i]
            if i < nParamsToPlot-1: 
                ticklabels = [item.get_text() for item in g.subplots[-1,i].xaxis.get_ticklabels()]
                ax.xaxis.set_ticklabels(ticklabels)
                ax.tick_params(labelbottom='off',labeltop='on')

            # Add shaded region showing 1D 68% credible interval
            y0, y1 = ax.get_ylim()
            lev = samples.get1DDensity(par_name).getLimits(settings['contours'][0])

            ax.add_patch(
                    Rectangle((lev[0], y0), 
                    lev[1]-lev[0], 
                    y1-y0, 
                    facecolor="grey", 
                    alpha=0.5)
                    )

            # Indicate the value of the "true" parameter
            if self.mock_catalogue is not None:
                name = names[i]
                value = self.mock_catalogue.get_param_values(ID, (name,))
                if "log" in self.adjust_params[name]:
                    if self.adjust_params[name]["log"]:
                        value = np.log10(value)

                ax.plot(value,
                        y0+(y1-y0)*0.05,
                        marker="D",
                        ms=8,
                        color="green") 

        if self.single_solutions is not None:
            row =  self.single_solutions['row'][self.single_solutions['ID']==ID]

            for i in range(nParamsToPlot):
                parX = keys[i]
                valueX = param_values[parX][row]

                if "log" in self.adjust_params[parX]:
                    if self.adjust_params[parX]["log"]:
                        valueX = np.log10(valueX)

                for j in range(nParamsToPlot):
                    ax = g.subplots[i,j]
                    if ax is None:
                        continue

                    if i == j:

                        ax.plot(valueX,
                                y0+(y1-y0)*0.05,
                                marker="*",
                                ms=12,
                                color="darkorange") 
                    else:

                        parY = keys[j]
                        valueY = param_values[parY][row]

                        if "log" in self.adjust_params[parY]:
                            if self.adjust_params[parY]["log"]:
                                valueY = np.log10(valueY)

                        ax.plot(valueY,
                                valueX,
                                marker="*",
                                ms=12,
                                color="darkorange") 

        if show:
            plt.show()
        else:
            # Now save the plot
            name = prepare_plot_saving(plot_name)
            g.export(name)

        plt.close()
        hdulist.close()


##                # Overplot the posterior median point
##                self.marginal.DrawPosteriorMedian(Appearance.PosteriorMedian)
##
##                # Overplot the posterior mean point
##                self.marginal.DrawPosteriorMean(Appearance.PosteriorMean)
##
##                # Overplot the maximum a posteriori
##                self.marginal.DrawMAP(Appearance.BestFit)
##
##                # Overplot the 1-D credible regions
##                self.marginal.DrawCredibleRegions(
##                    self.CredibleRegions)
##
##                # If has mock parameters, overplot the "true" values
##                try:
##                    mockValue = self.mockpar[iRow]
##                    if self.logPar[iRow]:
##                        mockValue = np.log10(mockValue)
##                    self.OneDim.PlotMockPar(
##                            mockValue,
##                            Appearance.MockPar)
##                except AttributeError:
##                    None
##
##                # Overplot the posterior mean point
##                self.joint.DrawPosteriorMean(
##                    Appearance.PosteriorMean)

        #plt.close(g.fig)



import numpy as np
import os
import logging
import cPickle

from beagle_utils import prepare_data_saving, BeagleDirectories, get_files_list

class MultiNestMode(object):

    def __init__(self, logEvidence, post_mean, max_likelihood, max_a_post):

        self.logEvidence = logEvidence
        self.post_mean = np.array(post_mean)
        self.max_likelihood = max_likelihood
        self.max_a_post = max_a_post


class MultiNestObject(object):

    def __init__(self, ID, logEvidence):

        self.ID = ID
        self.logEvidence = logEvidence
        self.mode = list()


    def add_mode(self, logEvidence, post_mean, max_likelihood, max_a_post):

        newMode = MultiNestMode( logEvidence, post_mean, max_likelihood,
                max_a_post )

        self.mode.append(newMode)


class MultiNestCatalogue(object):


    def load(self, file_name=None, n_par=None, file_list=None):
        """ 
        Load a 'MultiNest catalogue'.

        Parameters
        ----------
        file_name : str
            Name of the file containing the catalogue.

        Notes
        -----
        You then access the data for each object and each mode as
        self.MNObjects[0].mode[0].post_mean  ---> posterior mean values for
        the different parameters for object 0 and mode 0

        self.MNObjects[2].mode[1].max_likelihood ---> max_likelihood likelihood
        values for the different parameters for object 2 and mode 1
        """

        if file_name is None:
            try:
                tmp = BeagleDirectories.param_file 
                file_name = tmp.split('.')[-2] + '_MultiNest.cat'
            except:
                pass


        name = file_name
        if os.path.dirname(name) is None:
            name = os.path.join(BeagleDirectories.results_dir, 
                    BeagleDirectories.pypbeagle_data, 
                    file_name)

        if os.path.isfile(name):
            logging.info("Loading the `MultiNestCatalogue`: " + name)
            file = open(name, 'rb')
            self.MNObjects = cPickle.load(file)
            file.close()
            return

        name = prepare_data_saving(file_name)

        if n_par is not None:
            try:
                self.compute(n_par, file_list=file_list)
                return
            except:
                return


    def compute(self, n_par, file_list=None, file_name=None):
        """ 
        Compute a 'MultiNest catalogue'

        Parameters
        ----------
        n_par : int
            Number of free parameters in the BEAGLE run.

        file_list : iterable 
            Contains the list of MultiNest output files '*MNstats.dat'.

        file_name : str
            Name of the output 'MultiNest' catalogue.

        Notes
        -----
        You then access the data for each object and each mode as
        self.MNObjects[0].mode[0].post_mean  ---> posterior mean values for
        the different parameters for object 0 and mode 0

        self.MNObjects[2].mode[1].max_likelihood ---> max_likelihood likelihood
        values for the different parameters for object 2 and mode 1
        """

        if file_name is None:
            try:
                tmp = BeagleDirectories.param_file 
                file_name = tmp.split('.')[-2] + '_MultiNest.cat'
            except:
                pass

        if file_list is None:
            try:
                file_list, IDs = get_files_list(suffix=BeagleDirectories.MN_suffix)
            except:
                return

        self.MNObjects = list()

        # Loop over all files containing BEAGLE results
        for j, file in enumerate(file_list):

            # Object ID is what comes before the suffix (excluding the
            # directory tree!)
            end = file.find('_BEAGLE')
            objID = os.path.basename(file[:end]) 

            # Open the MultiNest "stats" file
            # Read the global evidence and total number of modes
            f = open(os.path.join(BeagleDirectories.results_dir, file), 'r')
            for i, line in enumerate(f):
                if i == 0:
                    logEvidence = float(line.split()[5])
                if i == 5:
                    n_modes = int(line.split()[3])
            f.close()

            # This number include the posterior mean, maximum likelihood and
            # maximum a posteriori for each parameter + the headers
            n_lines_per_mode = 8 + n_par*3

            # Useful information for the first mode start at line 11 (in Python
            # we count from 0)
            first_line = 10

            post_mean = list()
            max_likelihood = list()
            max_a_post = list()
            mode = list()
            
            # This object will contain the parameters (post mean, ecc...) for
            # each mode found by MultiNest
            MNObj = MultiNestObject(objID, logEvidence)

            # Now we read the evidence, post mean, maximum likelihood and map for each mode
            f = open(os.path.join(BeagleDirectories.results_dir, file), 'r')
            for i, line in enumerate(f):
                
                # Evidence
                if i == first_line:
                    logEv = float(line.split()[2])
                # Posterior mean for each parameter 
                elif( (i >= first_line+3) and (i < first_line+3+n_par) ):
                    post_mean.append( float(line.split()[1]) ) 
                # Maximum likelihood for each parameter 
                elif( (i >= first_line+6+n_par) and (i < first_line+6+2*n_par ) ):
                    max_likelihood.append( float(line.split()[1]) ) 
                # Maximum a posteriori for each parameter 
                elif( (i >= first_line+9+2*n_par) and (i < first_line+9+3*n_par ) ):
                    max_a_post.append( float(line.split()[1]) ) 

                # Once you've read the data for the first mode, put them into
                # the MultiNestObject!
                if i == (first_line + n_lines_per_mode):
                    MNObj.add_mode( logEv, post_mean, max_likelihood, max_a_post )
                    post_mean = list()
                    max_likelihood = list()
                    max_a_post = list()
                    first_line += n_lines_per_mode + 5

            f.close()

            # Finally append the MultiNestObject corresponding to the current
            # object to the catalogue
            self.MNObjects.append(MNObj)

        if file_name is not None:
            name = prepare_data_saving(file_name)

            fOut = open(name, 'w')

            # Print to a file the data read from the different *stats* files, and
            # saved into self.MNObjects  
            cPickle.dump(self.MNObjects, fOut, cPickle.HIGHEST_PROTOCOL)

            # Close the output file containing the cataogue
            fOut.close()
